@@ -77,7 +77,8 @@ This method is supposed to be overwritten by the corresponding token classes.
 """
 import logging
 import hashlib
-import datetime
+import traceback
+from datetime import datetime, timedelta
 
 from .error import (TokenAdminError,
                     ParameterError)
@@ -229,7 +230,9 @@ class TokenClass(object):
         
         An orphaned token means, that it has a user assigned, but the user 
         does not exist in the user store (anymore)
+
         :return: True / False
+        :rtype: bool
         """
         orphaned = False
         if self.token.first_owner:
@@ -297,6 +300,7 @@ class TokenClass(object):
     def set_tokeninfo(self, info):
         """
         Set the tokeninfo field in the DB. Old values will be deleted.
+
         :param info: dictionary with key and value
         :type info: dict
         :return:
@@ -316,6 +320,7 @@ class TokenClass(object):
     def add_tokeninfo(self, key, value, value_type=None):
         """
         Add a key and a value to the DB tokeninfo
+
         :param key:
         :param value:
         :return:
@@ -431,12 +436,13 @@ class TokenClass(object):
         :type options: dict
 
         :return: returns tuple of
-                 1. true or false for the pin match,
-                 2. the otpcounter (int) and the
-                 3. reply (dict) that will be added as
-                    additional information in the JSON response
-                    of ``/validate/check``.
-        :rtype: tuple
+
+            1. true or false for the pin match,
+            2. the otpcounter (int) and the
+            3. reply (dict) that will be added as additional information in
+                the JSON response of ``/validate/check``.
+
+        :rtype: tuple(bool, int, dict)
         """
         pin_match = False
         otp_counter = -1
@@ -452,7 +458,6 @@ class TokenClass(object):
 
         return pin_match, otp_counter, reply
 
-
     @staticmethod
     def decode_otpkey(otpkey, otpkeyformat):
         """
@@ -476,7 +481,6 @@ class TokenClass(object):
         else:
             raise ParameterError("Unknown OTP key format: {!r}".format(otpkeyformat))
 
-
     def update(self, param, reset_failcount=True):
         """
         Update the token object
@@ -674,6 +678,7 @@ class TokenClass(object):
     def set_realms(self, realms, add=False):
         """
         Set the list of the realms of a token.
+
         :param realms: realms the token should be assigned to
         :type realms: list
         :param add: if the realms should be added and not replaced
@@ -684,8 +689,9 @@ class TokenClass(object):
     def get_realms(self):
         """
         Return a list of realms the token is assigned to
+
         :return: realms
-        :rtype:l list
+        :rtype: list
         """
         return self.token.get_realms()
         
@@ -725,10 +731,11 @@ class TokenClass(object):
         """
         set the PIN of a token.
         Usually the pin is stored in a hashed way.
+
         :param pin: the pin to be set for the token
         :type pin: basestring
         :param encrypt: If set to True, the pin is stored encrypted and
-                        can be retrieved from the database again
+            can be retrieved from the database again
         :type encrypt: bool
         """
         storeHashed = not encrypt
@@ -791,7 +798,7 @@ class TokenClass(object):
             self.token.failcount = (self.token.failcount + 1)
             if self.token.failcount == self.token.maxfail:
                 self.add_tokeninfo(FAILCOUNTER_EXCEEDED,
-                                   datetime.datetime.now(tzlocal()).strftime(
+                                   datetime.now(tzlocal()).strftime(
                                        DATE_FORMAT))
         try:
             self.token.save()
@@ -821,6 +828,7 @@ class TokenClass(object):
     def get_hashlib(hLibStr):
         """
         Returns a hashlib function for a given string
+
         :param hLibStr: the hashlib
         :type hLibStr: string
         :return: the hashlib
@@ -859,7 +867,7 @@ class TokenClass(object):
         :param default: the default value, if the key does not exist
         :type default: string
         :return: the value for the key
-        :rtype: int or string
+        :rtype: int or str or dict
         """
         tokeninfo = self.token.get_info()
         if key:
@@ -879,6 +887,7 @@ class TokenClass(object):
         """
         Sets the counter for the maximum allowed successful logins
         as key "count_auth_success_max" in token info
+
         :param count: a number
         :type count: int
         """
@@ -889,6 +898,7 @@ class TokenClass(object):
         """
         Sets the counter for the occurred successful logins
         as key "count_auth_success" in token info
+
         :param count: a number
         :type count: int
         """
@@ -899,6 +909,7 @@ class TokenClass(object):
         """
         Sets the counter for the maximum allowed login attempts
         as key "count_auth_max" in token info
+
         :param count: a number
         :type count: int
         """
@@ -909,6 +920,7 @@ class TokenClass(object):
         """
         Sets the counter for the occurred login attepms
         as key "count_auth" in token info
+
         :param count: a number
         :type count: int
         """
@@ -946,8 +958,9 @@ class TokenClass(object):
         """
         returns the end of validity period (if set)
         if not set, "" is returned.
+
         :return: the end of the validity period
-        :rtype: string
+        :rtype: str
         """
         end = self.get_tokeninfo("validity_period_end", "")
         if end:
@@ -958,21 +971,30 @@ class TokenClass(object):
     def set_validity_period_end(self, end_date):
         """
         sets the end date of the validity period for a token
+
         :param end_date: the end date in the format YYYY-MM-DDTHH:MM+OOOO
                          if the format is wrong, the method will
                          throw an exception
-        :type end_date: string
+        :type end_date: str
         """
-        #  upper layer will catch. we just try to verify the date format
-        d = parse_date_string(end_date)
-        self.add_tokeninfo("validity_period_end", d.strftime(DATE_FORMAT))
+        if not end_date:
+            self.del_tokeninfo("validity_period_end")
+        else:
+            #  upper layer will catch. we just try to verify the date format
+            try:
+                d = parse_date_string(end_date)
+            except ValueError as _e:
+                log.debug('{0!s}'.format(traceback.format_exc()))
+                raise TokenAdminError('Could not parse validity period end date!')
+            self.add_tokeninfo("validity_period_end", d.strftime(DATE_FORMAT))
 
     def get_validity_period_start(self):
         """
         returns the start of validity period (if set)
         if not set, "" is returned.
+
         :return: the start of the validity period
-        :rtype: string
+        :rtype: str
         """
         start = self.get_tokeninfo("validity_period_start", "")
         if start:
@@ -983,20 +1005,28 @@ class TokenClass(object):
     def set_validity_period_start(self, start_date):
         """
         sets the start date of the validity period for a token
+
         :param start_date: the start date in the format YYYY-MM-DDTHH:MM+OOOO
                            if the format is wrong, the method will
                            throw an exception
-        :type start_date: string
+        :type start_date: str
         """
-        d = parse_date_string(start_date)
-        self.add_tokeninfo("validity_period_start", d.strftime(DATE_FORMAT))
+        if not start_date:
+            self.del_tokeninfo("validity_period_start")
+        else:
+            try:
+                d = parse_date_string(start_date)
+            except ValueError as _e:
+                log.debug('{0!s}'.format(traceback.format_exc()))
+                raise TokenAdminError('Could not parse validity period start date!')
+
+            self.add_tokeninfo("validity_period_start", d.strftime(DATE_FORMAT))
 
     def set_next_pin_change(self, diff=None, password=False):
         """
         Sets the timestamp for the next_pin_change. Provide a
         difference like 90d (90 days).
 
-        Either provider the
         :param diff: The time delta.
         :type diff: basestring
         :param password: Do no set next_pin_change but next_password_change
@@ -1006,25 +1036,23 @@ class TokenClass(object):
         key = "next_pin_change"
         if password:
             key = "next_password_change"
-        new_date = datetime.datetime.now(tzlocal()) + datetime.timedelta(days=days)
+        new_date = datetime.now(tzlocal()) + timedelta(days=days)
         self.add_tokeninfo(key, new_date.strftime(DATE_FORMAT))
 
     def is_pin_change(self, password=False):
         """
         Returns true if the pin of the token needs to be changed.
+
         :param password: Whether the password needs to be changed.
         :type password: bool
-
         :return: True or False
         """
         key = "next_pin_change"
         if password:
             key = "next_password_change"
         sdate = self.get_tokeninfo(key)
-        #date_change = datetime.datetime.strptime(sdate, DATE_FORMAT)
         date_change = parse_date_string(parse_legacy_time(sdate))
-        return datetime.datetime.now(tzlocal()) > date_change
-
+        return datetime.now(tzlocal()) > date_change
 
     @check_token_locked
     def inc_count_auth_success(self):
@@ -1040,6 +1068,13 @@ class TokenClass(object):
                              "count_auth": int(auth_counter)})
         return succcess_counter
 
+    @check_token_locked
+    def post_success(self):
+        """
+        Run anything after a token was used for successful authentication
+        """
+        return
+
     @check_token_locked
     def inc_count_auth(self):
         """
@@ -1066,11 +1101,11 @@ class TokenClass(object):
                         "failcounter_clear_timeout: "
                         "{0!s}".format(exx))
         if timeout and self.token.failcount == self.get_max_failcount():
-            now = datetime.datetime.now(tzlocal())
+            now = datetime.now(tzlocal())
             lastfail = self.get_tokeninfo(FAILCOUNTER_EXCEEDED)
             if lastfail is not None:
                 failcounter_exceeded = parse_legacy_time(lastfail, return_date=True)
-                if now > failcounter_exceeded + datetime.timedelta(minutes=timeout):
+                if now > failcounter_exceeded + timedelta(minutes=timeout):
                     self.reset()
                     return True
         return False
@@ -1079,7 +1114,9 @@ class TokenClass(object):
         """
         Checks if the failcounter is exceeded. It returns True, if the
         failcounter is less than maxfail
+
         :return: True or False
+        :rtype: bool
         """
         return self.token.failcount < self.token.maxfail
 
@@ -1106,7 +1143,7 @@ class TokenClass(object):
 
     def check_validity_period(self):
         """
-        This checks if the datetime.datetime.now() is within the validity
+        This checks if the datetime.now() is within the validity
         period of the token.
 
         :return: success
@@ -1116,15 +1153,13 @@ class TokenClass(object):
         end = self.get_validity_period_end()
 
         if start:
-            #dt_start = datetime.datetime.strptime(start, DATE_FORMAT)
             dt_start = parse_legacy_time(start, return_date=True)
-            if dt_start > datetime.datetime.now(tzlocal()):
+            if dt_start > datetime.now(tzlocal()):
                 return False
 
         if end:
-            #dt_end = datetime.datetime.strptime(end, DATE_FORMAT)
             dt_end = parse_legacy_time(end, return_date=True)
-            if dt_end < datetime.datetime.now(tzlocal()):
+            if dt_end < datetime.now(tzlocal()):
                 return False
 
         return True
@@ -1324,9 +1359,9 @@ class TokenClass(object):
         Each token type can decide on its own under which condition a challenge
         is triggered by overwriting this method.
 
-        **please note**: in case of pin policy == 2 (no pin is required)
-        the ``check_pin`` would always return true! Thus each request
-        containing a ``data`` or ``challenge`` would trigger a challenge!
+        .. note:: in case of ``pin policy == 2`` (no pin is required)
+                  the ``check_pin`` would always return true! Thus each request
+                  containing a ``data`` or ``challenge`` would trigger a challenge!
 
         The Challenge workflow is like this.
 
@@ -1334,15 +1369,19 @@ class TokenClass(object):
         a request which will create a new challenge (is_challenge_request) or if
         this is a response to an existing challenge (is_challenge_response).
         In these two cases during request processing the following functions are
-        called.
+        called::
 
-        is_challenge_request or is_challenge_response
-                 |                       |
-                 V                       V
-        create_challenge        check_challenge
-                 |                       |
-                 V                       V
-        challenge_janitor       challenge_janitor
+            is_challenge_request or is_challenge_response  <-------+
+                     |                       |                     |
+                     V                       V                     |
+            create_challenge        check_challenge_response     create_challenge
+                     |                       |                     ^
+                     |                       |                     |
+                     |              has_further_challenge [yes] ---+
+                     |                      [no]
+                     |                       |
+                     V                       V
+            challenge_janitor       challenge_janitor
 
         :param passw: password, which might be pin or pin+otp
         :type passw: string
@@ -1350,7 +1389,6 @@ class TokenClass(object):
         :type user: User object
         :param options: dictionary of additional request parameters
         :type options: dict
-
         :return: true or false
         :rtype: bool
         """
@@ -1364,8 +1402,8 @@ class TokenClass(object):
 
     def is_challenge_response(self, passw, user=None, options=None):
         """
-        This method checks, if this is a request, that is the response to
-        a previously sent challenge.
+        This method checks, if this is a request that is supposed to be
+        the answer to a previous challenge.
 
         The default behaviour to check if this is the response to a
         previous challenge is simply by checking if the request contains
@@ -1374,6 +1412,7 @@ class TokenClass(object):
 
         This method does not try to verify the response itself!
         It only determines, if this is a response for a challenge or not.
+        If the challenge still exists, is checked in has_db_challenge_response.
         The response is verified in check_challenge_response.
 
         :param passw: password, which might be pin or pin+otp
@@ -1386,11 +1425,28 @@ class TokenClass(object):
         :rtype: bool
         """
         options = options or {}
+        transaction_id = options.get("transaction_id") or options.get("state")
+        return bool(transaction_id)
+
+    def has_db_challenge_response(self, passw, user=None, options=None):
+        """
+        This method checks, if the given transaction_id is actually the response to
+        a real challenge. To do so, it verifies, if there is a DB entry for the
+        given serial number and transaction_id.
+        This is to avoid side effects by passing non-existent transaction_ids.
+
+        This method checks, if the token still has a challenge
+
+        :param passw:
+        :param user:
+        :param options:
+        :return:
+        """
+        options = options or {}
         challenge_response = False
         transaction_id = options.get("transaction_id") or options.get("state")
         if transaction_id:
-            # Now we also need to check, if the transaction_id is an entry to the
-            # serial number of this token
+            # Now we also need to check, if there is a corresponding DB entry
             chals = get_challenges(serial=self.token.serial, transaction_id=transaction_id)
             challenge_response = bool(chals)
 
@@ -1449,6 +1505,19 @@ class TokenClass(object):
         self.challenge_janitor()
         return otp_counter
 
+    # TODO: Add policy decorator like (challenge_response_allowed),
+    #  that does a PIN reset, if there is no further challenge
+    def has_further_challenge(self, options=None):
+        """
+        Returns true, if a token requires more than one challenge during challenge response
+        authentication. This could be a 4eyes token or indexed secret token, that queries more
+        than on input.
+
+        :param options: Additional options from the request
+        :return: True, if this very token requires further challenges
+        """
+        return False
+
     @staticmethod
     def challenge_janitor():
         """
@@ -1574,6 +1643,7 @@ class TokenClass(object):
         policies.
 
         The returned dictionary is added to the parameters of the API call.
+
         :param g: context object, see documentation of ``Match``
         :param params: The call parameters
         :type params: dict
@@ -1612,7 +1682,7 @@ class TokenClass(object):
                 # We need to set the timezone manually
                 last_success_auth = last_success_auth.replace(tzinfo=tzutc())
             # The last auth is to far in the past
-            if last_success_auth + tdelta < datetime.datetime.now(tzlocal()):
+            if last_success_auth + tdelta < datetime.now(tzlocal()):
                 res = False
                 log.debug("The last successful authentication is too old: "
                           "{0!s}".format(last_success_auth))
