@@ -6,6 +6,7 @@ This tests the files
 """
 import datetime
 import os
+import types
 
 from mock import mock
 
@@ -158,6 +159,18 @@ class AuditTestCase(MyTestCase):
         res = search(self.app.config, {"timelimit": "-1d"})
         self.assertTrue(res.get("count") == 0, res)
 
+        # test that hidden columns are not returned by audit search
+        # write some test data
+        self.Audit.log({"action": "/validate/check",
+                        "success": False})
+        self.Audit.finalize_log()
+        hidden_columns = [u"number", u"action_detail"]
+        res = search(self.app.config, {"hidden_columns": hidden_columns})
+        self.assertTrue(res.get("count") == 1, res)
+        # hidden columns are not returned in the auditdata
+        for col in hidden_columns:
+            self.assertTrue(col not in res.get("auditdata")[0], res)
+
     def test_04_lib_download(self):
         # Prepare some audit entries:
         self.Audit.log({"serial": "serial1"})
@@ -176,8 +189,7 @@ class AuditTestCase(MyTestCase):
         self.Audit.finalize_log()
 
         audit_log = self.Audit.csv_generator()
-        self.assertTrue(type(audit_log).__name__ == "generator",
-                        type(audit_log).__name__)
+        self.assertIsInstance(audit_log, types.GeneratorType)
 
         count = 0
         for audit_entry in audit_log:
@@ -306,6 +318,24 @@ class AuditTestCase(MyTestCase):
         self.assertEqual(audit_log.auditdata[0].get("token_type"), "spass")
 
 
+class AuditColumnLengthTestCase(OverrideConfigTestCase):
+    class Config(TestingConfig):
+        # this needs to exist on app creation
+        PI_AUDIT_SQL_COLUMN_LENGTH = {'user': 10}
+
+    def setUp(self):
+        self.Audit = getAudit(self.app.config)
+        self.Audit.clear()
+
+    def test_01_check_custom_column_lengths(self):
+        # Since changing the database during tests would be a hassle, we just
+        # reduce the custom length and check if it gets applied
+        self.Audit.log({"action": "test06", "tokentype": "spass",
+                        "user": "a_very_long_username@long_realm"})
+        self.Audit._truncate_data()
+        self.assertEqual(len(self.Audit.audit_data.get("user")), 10, self.Audit.audit_data)
+
+
 class AuditFileTestCase(OverrideConfigTestCase):
     class Config(TestingConfig):
         # this needs to exist on app creation
@@ -336,28 +366,39 @@ class AuditFileTestCase(OverrideConfigTestCase):
     @log_capture()
     def test_30_logger_audit_qualname(self, capture):
         # Check that the default qualname is 'privacyidea.lib.auditmodules.loggeraudit'
+        # The audit log runs 2 seconds - mocked
         current_utc_time = datetime.datetime(2018, 3, 4, 5, 6, 8)
-        with mock.patch('privacyidea.lib.auditmodules.loggeraudit.datetime') as mock_dt:
-            mock_dt.utcnow.return_value = current_utc_time
-            a = LoggerAudit(config={})
-            a.log({"action": "No PI_AUDIT_LOGGER_QUALNAME given"})
-            a.finalize_log()
-            capture.check_present(
-                ('privacyidea.lib.auditmodules.loggeraudit', 'INFO',
-                 '{{"action": "No PI_AUDIT_LOGGER_QUALNAME given", "policies": "", '
-                 '"timestamp": "{timestamp}"}}'.format(timestamp=current_utc_time.isoformat())))
+        startdate_time = datetime.datetime(2018, 3, 4, 5, 6, 6)
+        with mock.patch('privacyidea.lib.auditmodules.loggeraudit.datetime') as mock_timestamp:
+            with mock.patch('privacyidea.lib.auditmodules.base.datetime.datetime') as mock_startdate:
+                mock_timestamp.utcnow.return_value = current_utc_time
+                mock_startdate.now.return_value = startdate_time
+                a = LoggerAudit(config={})
+                a.log({"action": "No PI_AUDIT_LOGGER_QUALNAME given"})
+                a.finalize_log()
+                capture.check_present(
+                    ('privacyidea.lib.auditmodules.loggeraudit', 'INFO',
+                     '{{"action": "No PI_AUDIT_LOGGER_QUALNAME given", "duration": "0:00:02", '
+                     '"policies": "", "startdate": "{startdate}", '
+                     '"timestamp": "{timestamp}"}}'.format(timestamp=current_utc_time.isoformat(),
+                                                           startdate=startdate_time.isoformat())))
 
         # Now change the qualname to 'pi-audit'
         current_utc_time = datetime.datetime(2020, 3, 4, 5, 6, 8)
+        startdate_time = datetime.datetime(2020, 3, 4, 5, 6, 0)
         with mock.patch('privacyidea.lib.auditmodules.loggeraudit.datetime') as mock_dt:
-            mock_dt.utcnow.return_value = current_utc_time
-            a = LoggerAudit(config={"PI_AUDIT_LOGGER_QUALNAME": "pi-audit"})
-            a.log({"action": "PI_AUDIT_LOGGER_QUALNAME given"})
-            a.finalize_log()
-            capture.check_present(
-                ('pi-audit', 'INFO',
-                 '{{"action": "PI_AUDIT_LOGGER_QUALNAME given", "policies": "", '
-                 '"timestamp": "{timestamp}"}}'.format(timestamp=current_utc_time.isoformat())))
+            with mock.patch('privacyidea.lib.auditmodules.base.datetime.datetime') as mock_startdate:
+                mock_dt.utcnow.return_value = current_utc_time
+                mock_startdate.now.return_value = startdate_time
+                a = LoggerAudit(config={"PI_AUDIT_LOGGER_QUALNAME": "pi-audit"})
+                a.log({"action": "PI_AUDIT_LOGGER_QUALNAME given"})
+                a.finalize_log()
+                capture.check_present(
+                    ('pi-audit', 'INFO',
+                     '{{"action": "PI_AUDIT_LOGGER_QUALNAME given", "duration": "0:00:08", '
+                     '"policies": "", "startdate": "{startdate}", '
+                     '"timestamp": "{timestamp}"}}'.format(timestamp=current_utc_time.isoformat(),
+                                                           startdate=startdate_time.isoformat())))
 
 
 class ContainerAuditTestCase(OverrideConfigTestCase):
@@ -372,7 +413,8 @@ class ContainerAuditTestCase(OverrideConfigTestCase):
                                                          "privacyidea.lib.auditmodules.sqlaudit"],
                             "PI_AUDIT_CONTAINER_READ": "privacyidea.lib.auditmodules.sqlaudit",
                             "PI_AUDIT_NO_SIGN": True,
-                            "PI_AUDIT_SQL_URI": 'sqlite:///' + os.path.join(basedir, 'data-test.sqlite')})
+                            "PI_AUDIT_SQL_URI": 'sqlite:///' + os.path.join(basedir, 'data-test.sqlite')},
+                           startdate=None)
         self.assertFalse(a.has_data)
         a.log({"action": "something_test_30"})
         self.assertTrue(a.has_data)
@@ -406,17 +448,30 @@ class ContainerAuditTestCase(OverrideConfigTestCase):
                             "PI_AUDIT_NO_SIGN": True,
                             "PI_AUDIT_SQL_URI": 'sqlite:///' + os.path.join(basedir, 'data-test.sqlite')})
         a.log({"action": "something_test_35"})
+        a.add_to_log({'action_detail': 'some detail'})
+        a.add_policy('some policy')
         a.finalize_log()
         r = a.search({"action": "*something_test_35*"})
         # The search should go to the sql audit
         self.assertEqual(r.total, 1)
         self.assertEqual(r.auditdata[0].get("action"), u"something_test_35")
+        self.assertEqual(r.auditdata[0].get("action_detail"), u"some detail")
+        self.assertEqual(r.auditdata[0].get("policies"), u"some policy")
         # now check the log file
         with open("audit.log") as file:
             c = file.readlines()
             self.assertIn("something_test_35", c[-1])
+            self.assertIn("some detail", c[-1])
+            self.assertIn("some policy", c[-1])
         os.unlink('audit.log')
 
+        # check the CSV output
+        csv = a.csv_generator()
+        self.assertIsInstance(csv, types.GeneratorType)
+        csv_list = [c for c in csv]
+        self.assertGreater(len(csv_list), 0, csv_list)
+        self.assertTrue(any(['something_test_35' in l for l in csv_list]))
+
     def test_20_container_audit_wrong_module(self):
         # Test what happens with a non-existing module
         import os
